/***************************************
  $Header: /home/amb/wwwoffle/RCS/misc.c 2.55 2001/04/10 18:08:37 amb Exp $

  WWWOFFLE - World Wide Web Offline Explorer - Version 2.6c.
  Miscellaneous HTTP / HTML functions.
  ******************/ /******************
  Written by Andrew M. Bishop

  This file Copyright 1997,98,99,2000,01 Andrew M. Bishop
  It may be distributed under the GNU Public License, version 2, or
  any higher version.  See section COPYING of the GNU Public license
  for conditions under which this file may be redistributed.
  ***************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "misc.h"
#include "errors.h"
#include "proto.h"
#include "config.h"
#include "md5.h"


/*++++++++++++++++++++++++++++++++++++++
  Split a URL into a protocol, hostname, path name and an argument list.

  URL *SplitURL Returns a URL structure containing the information.

  char *url The name of the url to split.
  ++++++++++++++++++++++++++++++++++++++*/

URL *SplitURL(char *url)
{
 URL *Url=(URL*)malloc(sizeof(URL));
 char *copyurl,*mallocurl=malloc(strlen(url)+2);
 int i=0,n=0;
 char *colon,*slash,*at,*ques,*semi,*temppath,root[2];

 copyurl=mallocurl;
 strcpy(copyurl,url);

 /* Protocol */

 colon=strchr(copyurl,':');
 slash=strchr(copyurl,'/');
 at   =strchr(copyurl,'@');

 if(slash==copyurl)                     /* /dir/... (local) */
   {
    Url->proto=(char*)malloc(5);
    strcpy(Url->proto,"http");
   }
 else if(colon && slash && (colon+1)==slash) /* http://... */
   {
    *colon=0;
    Url->proto=(char*)malloc(colon-copyurl+1);
    strcpy(Url->proto,copyurl);
    copyurl=slash+1;
    if(*copyurl=='/')
       copyurl++;

    colon=strchr(copyurl,':');
    slash=strchr(copyurl,'/');
   }
 else if(colon && !isdigit(*(colon+1)) &&
         (!slash || colon<slash) &&
         (!at || (slash && at>slash)))  /* http:www.foo.com/...[:@]... */
   {
    *colon=0;
    Url->proto=(char*)malloc(colon-copyurl+1);
    strcpy(Url->proto,copyurl);
    copyurl=colon+1;

    colon=strchr(copyurl,':');
   }
 else                                   /* www.foo.com:80/... */
   {
    Url->proto=(char*)malloc(5);
    strcpy(Url->proto,"http");
   }

 for(i=0;Url->proto[i];i++)
    Url->proto[i]=tolower(Url->proto[i]);

 Url->Protocol=NULL;

 for(i=0;i<NProtocols;i++)
    if(!strcmp(Protocols[i].name,Url->proto))
       Url->Protocol=&Protocols[i];

 /* Password */

 if(at && at>copyurl && (!slash || slash>at))
   {
    char *at2;

    if(colon && at>colon)               /* user:pass@www.foo.com...[/]... */
      {
       *colon=0;
       Url->user=URLDecodeGeneric(copyurl);
       *at=0;
       Url->pass=URLDecodeGeneric(colon+1);
       copyurl=at+1;
      }
    else if(colon && (at2=strchr(at+1,'@')) && /* user@host:pass@www.foo.com...[/]... */
            at2>colon && at2<slash)            /* [not actually valid, but common]    */
      {
       *colon=0;
       Url->user=URLDecodeGeneric(copyurl);
       *at2=0;
       Url->pass=URLDecodeGeneric(colon+1);
       copyurl=at2+1;
      }
    else                               /* user@www.foo.com...[:/]... */
      {
       *at=0;
       Url->user=URLDecodeGeneric(copyurl);
       Url->pass=NULL;
       copyurl=at+1;
      }
   }
 else
   {
    if(at==copyurl)             /* @www.foo.com... */
       copyurl++;

    Url->user=NULL;
    Url->pass=NULL;
   }

 /* Parameters */

 semi=strchr(copyurl,';');
 ques=strchr(copyurl,'?');

 Url->params=NULL;

 if(semi && (!ques || semi<ques)) /* ../path;...[?]... */
   {
    *semi++=0;

    if(ques)
       *ques=0;

    if(*semi)
       Url->params=URLRecodeFormArgs(semi);
   }

 /* Arguments */

 if(ques)                       /* ../path?... */
   {
    *ques++=0;
    Url->args=URLRecodeFormArgs(ques);
   }
 else
    Url->args=NULL;

 /* Hostname */

 if(*copyurl=='/')              /* /path/... (local) */
   {
    Url->host=GetLocalHost(1);
    Url->local=1;
   }
 else                           /* www.foo.com... */
   {
    Url->host=copyurl;
    Url->local=0;

    if(slash && (!ques || slash<ques)) /* www.foo.com/...[?]... */
       copyurl=slash;
    else                        /* www.foo.com[?]... */
      {root[0]='/';root[1]=0;copyurl=root;}
   }

 /* Pathname */

 temppath=URLDecodeGeneric(copyurl);
 Url->path=URLEncodePath(temppath);
 free(temppath);

 /* Hostname (cont) */

 if(!Url->local)
   {
    *copyurl=0;
    copyurl=Url->host;
    Url->host=(char*)malloc(strlen(copyurl)+1);
    strcpy(Url->host,copyurl);
   }

 for(i=0;Url->host[i] && Url->host[i]!=':';i++)
    Url->host[i]=tolower(Url->host[i]);

 if(Url->host[i]==':')
    if(atoi(&Url->host[i+1])==(Url->Protocol?Url->Protocol->defport:80))
       Url->host[i]=0;

 if(!Url->local && IsLocalHost(Url->host,1) && Url->Protocol && Url->Protocol==&Protocols[0])
   {
    free(Url->host);
    Url->host=GetLocalHost(1);
    Url->local=1;
   }

 /* Canonicalise the URL. */

 Url->name=(char*)malloc(strlen(Url->proto)+
                         strlen(Url->host)+
                         strlen(Url->path)+
                         (Url->params?strlen(Url->params):0)+
                         (Url->args?strlen(Url->args):0)+
                         8);

 strcpy(Url->name,Url->proto);
 n=strlen(Url->proto);
 strcpy(Url->name+n,"://");
 n+=3;

 Url->hostp=Url->name+n;

 strcpy(Url->name+n,Url->host);
 n+=strlen(Url->host);

 Url->pathp=Url->name+n;

 strcpy(Url->name+n,Url->path);
 n+=strlen(Url->path);

 if(Url->params && *Url->params)
   {
    strcpy(Url->name+n,";");
    strcpy(Url->name+n+1,Url->params);
    n+=strlen(Url->params)+1;
   }

 if(Url->args && *Url->args)
   {
    strcpy(Url->name+n,"?");
    strcpy(Url->name+n+1,Url->args);
   }

 if(Url->user)
   {
    char *encuserpass;

    Url->file=(char*)malloc(strlen(Url->name)+
                            3*strlen(Url->user)+
                            (Url->pass?3*strlen(Url->pass):0)+
                            8);

    n=Url->hostp-Url->name;
    strncpy(Url->file,Url->name,n);

    encuserpass=URLEncodePassword(Url->user);

    strcpy(Url->file+n,encuserpass);
    n+=strlen(encuserpass);

    free(encuserpass);

    if(Url->pass)
      {
       encuserpass=URLEncodePassword(Url->pass);

       strcpy(Url->file+n,":");
       strcpy(Url->file+n+1,encuserpass);
       n+=strlen(encuserpass)+1;

       free(encuserpass);
      }

    strcpy(Url->file+n,"@");
    strcpy(Url->file+n+1,Url->hostp);
   }
 else
    Url->file=Url->name;

 if(Url->Protocol && !Url->Protocol->proxyable)
   {
    char *localhost=GetLocalHost(1);
    Url->link=(char*)malloc(strlen(Url->name)+strlen(localhost)+8);
    sprintf(Url->link,"http://%s/%s/%s",localhost,Url->proto,Url->hostp);
   }
 else
    Url->link=Url->name;

 Url->dir=Url->host;

#if defined(__CYGWIN__)
 if(strchr(Url->host,':'))
   {
    char *colon=strchr(Url->host,':');
    Url->dir=(char*)malloc(strlen(Url->host));
    strcpy(Url->dir,Url->host);

    Url->dir[colon-Url->host]='!';
   }
#endif

 /* end */

 free(mallocurl);

 return(Url);
}


/*++++++++++++++++++++++++++++++++++++++
  Add a password to an existing URL.

  URL *Url The URL to add the username and password to.

  char *user The username.

  char *pass The password.
  ++++++++++++++++++++++++++++++++++++++*/

void AddURLPassword(URL *Url,char *user,char *pass)
{
 int n;
 char *encuserpass;

 Url->user=(char*)malloc(strlen(user)+1);
 strcpy(Url->user,user);

 if(pass)
   {
    Url->pass=(char*)malloc(strlen(pass)+1);
    strcpy(Url->pass,pass);
   }

 Url->file=(char*)malloc(strlen(Url->name)+
                         3*strlen(Url->user)+
                         (Url->pass?3*strlen(Url->pass):0)+8);

 n=Url->hostp-Url->name;
 strncpy(Url->file,Url->name,n);

 encuserpass=URLEncodePassword(Url->user);

 strcpy(Url->file+n,encuserpass);
 n+=strlen(encuserpass);

 free(encuserpass);

 if(Url->pass)
   {
    encuserpass=URLEncodePassword(Url->pass);

    strcpy(Url->file+n,":");
    strcpy(Url->file+n+1,encuserpass);
    n+=strlen(encuserpass)+1;

    free(encuserpass);
   }

 strcpy(Url->file+n,"@");
 strcpy(Url->file+n+1,Url->hostp);
}


/*++++++++++++++++++++++++++++++++++++++
  Free the memory in a URL.

  URL *Url The URL to free.
  ++++++++++++++++++++++++++++++++++++++*/

void FreeURL(URL *Url)
{
 if(Url->name!=Url->link)
    free(Url->link);

 if(Url->name!=Url->file)
    free(Url->file);

 if(Url->dir!=Url->host)
    free(Url->dir);

 free(Url->name);

 if(Url->proto)  free(Url->proto);
 if(Url->host)   free(Url->host);
 if(Url->path)   free(Url->path);
 if(Url->params) free(Url->params);
 if(Url->args)   free(Url->args);

 if(Url->user)   free(Url->user);
 if(Url->pass)   free(Url->pass);

 free(Url);
}


/*++++++++++++++++++++++++++++++++++++++
  Convert a url reference from a page to an absolute one.

  char *LinkURL Return the linked to URL or the original link if unchanged.

  URL *Url The page that we are looking at.

  char *link The link from the page.
  ++++++++++++++++++++++++++++++++++++++*/

char *LinkURL(URL *Url,char *link)
{
 char *new=link;
 char *colon=strchr(link,':');
 char *slash=strchr(link,'/');

 if(colon && slash && colon<slash)
    ;
 else if(!*link)
   {
    new=(char*)malloc(strlen(Url->name)+1);
    strcpy(new,Url->name);
   }
 else
   {
    if(*link=='/')
      {
       new=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+strlen(link)+8);
       sprintf(new,"%s://%s%s",Url->proto,Url->host,link);
      }
    else
      {
       int j;
       char *path=(char*)malloc(strlen(Url->path)+strlen(link)+2);

       strcpy(path,Url->path);

       for(j=strlen(path);j>0;j--)
          if(path[j]=='/')
             break;
       path[j]=0;

       strcat(path,"/");
       strcat(path,link);

       CanonicaliseName(path);

       new=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+strlen(path)+8);
       sprintf(new,"%s://%s%s",Url->proto,Url->host,path);
       free(path);
      }
   }

 return(new);
}


/*++++++++++++++++++++++++++++++++++++++
  Canonicalise a file name by removing '/../', '/./' and '//' references.

  char *CanonicaliseName Returns the argument modified.

  char *name The original name
  ++++++++++++++++++++++++++++++++++++++*/

char *CanonicaliseName(char *name)
{
 char *match,*name2;

 match=name;
 while((match=strstr(match,"/./")) || !strncmp(match=name,"./",2))
   {
    char *prev=match, *next=match+2;
    while((*prev++=*next++));
   }

 match=name;
 while((match=strstr(match,"//")))
   {
    char *prev=match, *next=match+1;
    while((*prev++=*next++));
   }

 match=name2=name;
 while((match=strstr(match,"/../")))
   {
    char *prev=match, *next=match+4;
    if((prev-name2)==2 && !strncmp(name2,"../",3))
      {name2+=3;match++;continue;}
    while(prev>name2 && *--prev!='/');
    match=prev;
    if(*prev=='/')prev++;
    while((*prev++=*next++));
   }

 match=&name[strlen(name)-2];
 if(match>=name && !strcmp(match,"/."))
    *match=0;

 match=&name[strlen(name)-3];
 if(match>=name && !strcmp(match,"/.."))
   {
    if(match==name)
       *++match=0;
    else
       while(match>name && *--match!='/')
          *match=0;
   }

#if 0 /* as used in cxref */

 match=&name[strlen(name)-1];
 if(match>name && !strcmp(match,"/"))
    *match=0;

 if(!*name)
    *name='.',*(name+1)=0;

#else /* as used in wwwoffle */

 if(!*name || !strncmp(name,"../",3))
    *name='/',*(name+1)=0;

#endif

 return(name);
}


/* To understand why the URLDecode*() and URLEncode*() functions are coded this way see README.URL */

/*+ For conversion from integer to hex string. +*/
static char hexstring[16]="0123456789ABCDEF";

/*+ For conversion from hex string to integer. +*/
static char unhexstring[256]={ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x00-0x0f "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x10-0x1f "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x20-0x2f " !"#$%&'()*+,-./" */
                               0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 0, 0, 0, 0, 0,  /* 0x30-0x3f "0123456789:;<=>?" */
                               0,10,11,12,13,14,15, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x40-0x4f "@ABCDEFGHIJKLMNO" */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x50-0x5f "PQRSTUVWXYZ[\]^_" */
                               0,10,11,12,13,14,15, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x60-0x6f "`abcdefghijklmno" */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x70-0x7f "pqrstuvwxyz{|}~ " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x80-0x8f "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0x90-0x9f "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0xa0-0xaf "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0xb0-0xbf "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0xc0-0xcf "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0xd0-0xdf "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 0xe0-0xef "                " */
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}; /* 0xf0-0xff "                " */


/*++++++++++++++++++++++++++++++++++++++
  Decode a string that has been UrlEncoded.

  char *URLDecodeGeneric Returns a malloced copy of the decoded string.

  const char *str The generic string to be decoded.
  ++++++++++++++++++++++++++++++++++++++*/

char *URLDecodeGeneric(const char *str)
{
 int i,j;
 char *copy=(char*)malloc(strlen(str)+1);

 for(i=0,j=0;str[i];i++)
    if(str[i]=='%' && str[i+1]!='\0' && str[i+2]!='\0')
      {
       unsigned int val=0;
       val=16*unhexstring[(unsigned char)str[++i]];
       val+=unhexstring[(unsigned char)str[++i]];
       copy[j++]=val;
      }
    else
       copy[j++]=str[i];

 copy[j]=0;

 return(copy);
}


/*++++++++++++++++++++++++++++++++++++++
  Decode a POSTed form or URL arguments that has been UrlEncoded.

  char *URLDecodeForm Returns a malloced copy of the decoded form data string.

  const char *str The form data string to be decoded.
  ++++++++++++++++++++++++++++++++++++++*/

char *URLDecodeFormArgs(const char *str)
{
 int i,j;
 char *copy=(char*)malloc(strlen(str)+1);

 /* This is the same as the generic function except that '+' is used for ' ' instead of '%20'. */

 for(i=0,j=0;str[i];i++)
    if(str[i]=='%' && str[i+1]!='\0' && str[i+2]!='\0')
      {
       unsigned int val=0;
       val=16*unhexstring[(unsigned char)str[++i]];
       val+=unhexstring[(unsigned char)str[++i]];
       copy[j++]=val;
      }
    else if(str[i]=='+')
       copy[j++]=' ';
    else
       copy[j++]=str[i];

 copy[j]=0;

 return(copy);
}


/*++++++++++++++++++++++++++++++++++++++
  Take a POSTed form data or a URL argument string and URLDecode and URLEncode it again.

  char *URLRecodeFormArgs Returns a malloced copy of the decoded and re-encoded form/argument string.

  const char *str The form/argument string to be decoded and re-encoded.
  ++++++++++++++++++++++++++++++++++++++*/

char *URLRecodeFormArgs(const char *str)
{
 int i,j;
 char *copy=(char*)malloc(3*strlen(str)+1);

 /*
   The characters in the range 0x00-0x1f and 0x7f-0xff are always disallowed.
   The '%' character is always disallowed because it is the quote character.
   RFC 1738 section 2.2 calls " <>"#%{}|\^~[]`" unsafe characters, I make an exception for "|~".
   RFC 1738 section 2.2 calls ";/?:@=&" reserved characters, I make an exception for "/:".
   RFC 1866 section 8.2.1 says that ' ' is replaced by '+'.
   I disallow "'" because it may lead to confusion.
   The unencoded characters "&=;" on the input are left unencoded on the output
   The encoded character "+" on the input is left encoded on the output
   The unencoded character "?" on the input is left unencoded on the output to handle broken servers.
 */

 static char allowed[256]=
 "                                "  /* 0x00-0x1f "                                " */
 " !  $   ()* ,-./0123456789:     "  /* 0x20-0x3f " !"#$%&'()*+,-./0123456789:;<=>?" */
 " ABCDEFGHIJKLMNOPQRSTUVWXYZ    _"  /* 0x40-0x5f "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_" */
 " abcdefghijklmnopqrstuvwxyz | ~ "  /* 0x60-0x7f "`abcdefghijklmnopqrstuvwxyz{|}~ " */
 "                                "  /* 0x80-0x9f "                                " */
 "                                "  /* 0xa0-0xbf "                                " */
 "                                "  /* 0xc0-0xdf "                                " */
 "                                "; /* 0xe0-0xff "                                " */

 for(i=0,j=0;str[i];i++)
    if(allowed[(unsigned char)str[i]]!=' ' || str[i]=='&' || str[i]=='=' || str[i]==';' || str[i]=='+' || str[i]=='?')
       copy[j++]=str[i];
    else if(str[i]=='%' && str[i+1]!='\0' && str[i+2]!='\0')
      {
       unsigned int val=0;
       val=16*unhexstring[(unsigned char)str[++i]];
       val+=unhexstring[(unsigned char)str[++i]];

       if(allowed[val]!=' ')
          copy[j++]=val;
       else if(val==' ')
          copy[j++]='+';
       else
         {
          copy[j++]='%';
          copy[j++]=toupper(str[i-1]);
          copy[j++]=toupper(str[i]);
         }
      }
    else if(str[i]==' ')
       copy[j++]='+';
    else
      {
       unsigned int val=str[i];
       copy[j]='%';
       copy[j+2]=hexstring[val%16];
       val>>=4;
       copy[j+1]=hexstring[val%16];
       j+=3;
      }

 copy[j]=0;

 return(copy);
}


/*++++++++++++++++++++++++++++++++++++++
  Encode a string suitable for use as a pathname.

  char *URLEncodePath Returns a malloced copy of the encoded pathname string.

  const char *str The pathname string to be encoded.
  ++++++++++++++++++++++++++++++++++++++*/

char *URLEncodePath(const char *str)
{
 int i,j;
 char *copy=(char*)malloc(3*strlen(str)+1);

 /*
   The characters in the range 0x00-0x1f and 0x7f-0xff are always disallowed.
   The '%' character is always disallowed because it is the quote character.
   RFC 1738 section 2.2 calls " <>"#%{}|\^~[]`" unsafe characters, I make an exception for '~'.
   RFC 1738 section 2.2 calls ";/?:@=&" reserved characters, I make an exception for "/:=".
   I disallow "'" because it may lead to confusion.
 */

 static char allowed[256]=
 "                                "  /* 0x00-0x1f "                                " */
 " !  $   ()*+,-./0123456789:  =  "  /* 0x20-0x3f " !"#$%&'()*+,-./0123456789:;<=>?" */
 " ABCDEFGHIJKLMNOPQRSTUVWXYZ    _"  /* 0x40-0x5f "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_" */
 " abcdefghijklmnopqrstuvwxyz   ~ "  /* 0x60-0x7f "`abcdefghijklmnopqrstuvwxyz{|}~ " */
 "                                "  /* 0x80-0x9f "                                " */
 "                                "  /* 0xa0-0xbf "                                " */
 "                                "  /* 0xc0-0xdf "                                " */
 "                                "; /* 0xe0-0xff "                                " */

 for(i=0,j=0;str[i];i++)
    if(allowed[(unsigned char)str[i]]!=' ')
       copy[j++]=str[i];
    else
      {
       unsigned int val=str[i];
       copy[j]='%';
       copy[j+2]=hexstring[val%16];
       val>>=4;
       copy[j+1]=hexstring[val%16];
       j+=3;
      }

 copy[j]=0;

 return(copy);
}


/*++++++++++++++++++++++++++++++++++++++
  Encode a string suitable for use with the POST method or as URL arguments.

  char *URLEncodeFormArgs Returns a malloced copy of the encoded form data string.

  const char *str The form data string to be encoded.
  ++++++++++++++++++++++++++++++++++++++*/

char *URLEncodeFormArgs(const char *str)
{
 int i,j;
 char *copy=(char*)malloc(3*strlen(str)+1);

 /*
   The characters in the range 0x00-0x1f and 0x7f-0xff are always disallowed.
   The '%' character is always disallowed because it is the quote character.
   RFC 1738 section 2.2 calls " <>"#%{}|\^~[]`" unsafe characters, I make an exception for "|~".
   RFC 1738 section 2.2 calls ";/?:@=&" reserved characters, I make an exception for "/:".
   RFC 1866 section 8.2.1 says that ' ' is replaced by '+'.
   I disallow ""'\`" because they may lead to confusion.
 */

 static char allowed[256]=
 "                                "  /* 0x00-0x1f "                                " */
 " !  $   ()* ,-./0123456789:     "  /* 0x20-0x3f " !"#$%&'()*+,-./0123456789:;<=>?" */
 " ABCDEFGHIJKLMNOPQRSTUVWXYZ    _"  /* 0x40-0x5f "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_" */
 " abcdefghijklmnopqrstuvwxyz | ~ "  /* 0x60-0x7f "`abcdefghijklmnopqrstuvwxyz{|}~ " */
 "                                "  /* 0x80-0x9f "                                " */
 "                                "  /* 0xa0-0xbf "                                " */
 "                                "  /* 0xc0-0xdf "                                " */
 "                                "; /* 0xe0-0xff "                                " */

 for(i=0,j=0;str[i];i++)
    if(allowed[(unsigned char)str[i]]!=' ')
       copy[j++]=str[i];
    else if(str[i]==' ')
       copy[j++]='+';
    else
      {
       unsigned int val=str[i];
       copy[j]='%';
       copy[j+2]=hexstring[val%16];
       val>>=4;
       copy[j+1]=hexstring[val%16];
       j+=3;
      }

 copy[j]=0;

 return(copy);
}


/*++++++++++++++++++++++++++++++++++++++
  Encode a string suitable for use as a username / password in a URL.

  char *URLEncodePassword Returns a malloced copy of the encoded username / password string.

  const char *str The password / username string to be encoded.
  ++++++++++++++++++++++++++++++++++++++*/

char *URLEncodePassword(const char *str)
{
 int i,j;
 char *copy=(char*)malloc(3*strlen(str)+1);

 /*
   The characters in the range 0x00-0x1f and 0x7f-0xff are always disallowed.
   The '%' character is always disallowed because it is the quote character.
   RFC 1738 section 2.2 calls " <>"#%{}|\^~[]`" unsafe characters, I make an exception for '~'.
   RFC 1738 section 2.2 calls ";/?:@=&" reserved characters, I make an exception for "=".
   RFC 1738 section 3.1 says that "@:/" are disallowed.
   I disallow "'()" because they may lead to confusion.
 */

 static char allowed[256]=
 "                                "  /* 0x00-0x1f "                                " */
 " !  $     *+,-. 0123456789   =  "  /* 0x20-0x3f " !"#$%&'()*+,-./0123456789:;<=>?" */
 " ABCDEFGHIJKLMNOPQRSTUVWXYZ    _"  /* 0x40-0x5f "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_" */
 " abcdefghijklmnopqrstuvwxyz   ~ "  /* 0x60-0x7f "`abcdefghijklmnopqrstuvwxyz{|}~ " */
 "                                "  /* 0x80-0x9f "                                " */
 "                                "  /* 0xa0-0xbf "                                " */
 "                                "  /* 0xc0-0xdf "                                " */
 "                                "; /* 0xe0-0xff "                                " */

 for(i=0,j=0;str[i];i++)
    if(allowed[(unsigned char)str[i]]!=' ')
       copy[j++]=str[i];
    else
      {
       unsigned int val=str[i];
       copy[j]='%';
       copy[j+2]=hexstring[val%16];
       val>>=4;
       copy[j+1]=hexstring[val%16];
       j+=3;
      }

 copy[j]=0;

 return(copy);
}


/*++++++++++++++++++++++++++++++++++++++
  Split a form body or URL arguments up into the component parts.

  char **SplitFormArgs Returns an array of pointers into a copy of the string.

  char *str The form data or arguments to split.
  ++++++++++++++++++++++++++++++++++++++*/

char **SplitFormArgs(char *str)
{
 char **args=(char**)malloc(9*sizeof(char*));
 char *p;
 int i=0;

 args[0]=(char*)malloc(strlen(str)+1);
 strcpy(args[0],str);

 p=args[0];
 while(*p)
   {
    if(*p=='&' || *p==';')
      {
       *p=0;
       if(i && !(i%8))
          args=(char**)realloc((void*)args,(i+10)*sizeof(char*));
       args[++i]=p+1;
      }
    p++;
   }

 args[++i]=NULL;

 return(args);
}


/*++++++++++++++++++++++++++++++++++++++
  Generate a hash value for a string.

  char *MakeHash Returns a string that can be used as the hashed string.

  const char *args The arguments.
  ++++++++++++++++++++++++++++++++++++++*/

char *MakeHash(const char *args)
{
 char md5[17];
 char *hash,*p;
 struct MD5Context ctx;

 /* Initialize the computation context.  */
 MD5Init (&ctx);

 /* Process whole buffer but last len % 64 bytes.  */
 MD5Update (&ctx, (const unsigned char*)args, strlen(args));

 /* Put result in desired memory area.  */
 MD5Final ((unsigned char *)md5, &ctx);

 md5[17]=0;

 hash=Base64Encode(md5,16);

 for(p=hash;*p;p++)
    if(*p=='/')
       *p='-';
    else if(*p=='=')
       *p=0;

 return(hash);
}


/*+ Conversion from time_t to date string and back +*/
static char *weekdays[7]={"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
static char *months[12]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};


/*++++++++++++++++++++++++++++++++++++++
  Convert the time into an RFC 822 compliant date.

  char *RFC822Date Returns a pointer to a fixed string containing the date.

  long t The time.

  int utc Set to true to get Universal Time, else localtime.
  ++++++++++++++++++++++++++++++++++++++*/

char *RFC822Date(long t,int utc)
{
 static char value[2][32];
 static int which=0;
 char weekday[4];
 char month[4];
 struct tm *tim;

 if(utc) /* Get UTC using English language. */
   {
    tim=gmtime(&t);

    strcpy(weekday,weekdays[tim->tm_wday]);
    strcpy(month,months[tim->tm_mon]);
   }
 else /* Get the local time using current language. */
   {
    tim=localtime(&t);
    if(tim->tm_isdst<0)
      {tim=gmtime(&t);utc=1;}

    strftime(weekday,4,"%a",tim);
    strftime(month,4,"%b",tim);
   }

 which^=1;

 /* Sun, 06 Nov 1994 08:49:37 GMT    ; RFC 822, updated by RFC 1123 */

 sprintf(value[which],"%3s, %02d %3s %4d %02d:%02d:%02d %s",
         weekday,
         tim->tm_mday,
         month,
         tim->tm_year+1900,
         tim->tm_hour,
         tim->tm_min,
         tim->tm_sec,
#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
         utc?"GMT":tim->tm_zone
#elif defined(__CYGWIN__)
         utc?"GMT":_tzname[tim->tm_isdst>0]
#else
         utc?"GMT":tzname[tim->tm_isdst>0]
#endif
         );

 return(value[which]);
}


/*++++++++++++++++++++++++++++++++++++++
  Convert a string representing a date into a time.

  long DateToTimeT Returns the time.

  const char *date The date string.
  ++++++++++++++++++++++++++++++++++++++*/

long DateToTimeT(const char *date)
{
 int  year,day,hour,min,sec;
 char monthstr[16];
 long retval=0;

 if(sscanf(date,"%*s %d %s %d %d:%d:%d",&day,monthstr,&year,&hour,&min,&sec)==6 ||
    sscanf(date,"%*s %d-%3s-%d %d:%d:%d",&day,monthstr,&year,&hour,&min,&sec)==6 ||
    sscanf(date,"%*s %3s %d %d:%d:%d %d",monthstr,&day,&hour,&min,&sec,&year)==6)
   {
    struct tm tim;
    int mon;

    for(mon=0;mon<12;mon++)
       if(!strcmp(monthstr,months[mon]))
          break;

    tim.tm_sec=sec;
    tim.tm_min=min;
    tim.tm_hour=hour;
    tim.tm_mday=day;
    tim.tm_mon=mon;
    if(year<38)
       tim.tm_year=year+100;
    else if(year<100)
       tim.tm_year=year;
    else
       tim.tm_year=year-1900;
    tim.tm_isdst=0;

    retval=mktime(&tim);

    if(retval==-1)
       retval=0;
   }
 else if(sscanf(date,"%ld %1s",&retval,monthstr)==1)
    ;

 return(retval);
}


/*+ The conversion from a 6 bit value to an ASCII character. +*/
static char base64[64]={'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
                        'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
                        'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
                        'w','x','y','z','0','1','2','3','4','5','6','7','8','9','+','/'};

/*++++++++++++++++++++++++++++++++++++++
  Decode a base 64 string.

  char *Base64Decode Return a malloced string containing the decoded version.

  const char *str The string to be decoded.

  int *l Returns the length of the decoded string.
  ++++++++++++++++++++++++++++++++++++++*/

char *Base64Decode(const char *str,int *l)
{
 int le=strlen(str);
 char *decoded=(char*)malloc(le+1);
 int i,j,k;

 while(str[le-1]=='=')
    le--;

 *l=3*(le/4)+(le%4)-1+!(le%4);

 for(j=0;j<le;j++)
    for(k=0;k<64;k++)
       if(base64[k]==str[j])
         {decoded[j]=k;break;}

 for(i=j=0;j<(le+4);i+=3,j+=4)
   {
    unsigned long s=0;

    for(k=0;k<4;k++)
       if((j+k)<le)
          s|=((unsigned long)decoded[j+k]&0xff)<<(18-6*k);

    for(k=0;k<3;k++)
       if((i+k)<*l)
          decoded[i+k]=(char)((s>>(16-8*k))&0xff);
   }
 decoded[*l]=0;

 return(decoded);
}


/*++++++++++++++++++++++++++++++++++++++
  Encode a string into base 64.

  char *Base64Encode Return a malloced string containing the encoded version.

  const char *str The string to be encoded.

  int l The length of the string to be encoded.
  ++++++++++++++++++++++++++++++++++++++*/

char *Base64Encode(const char *str,int l)
{
 int le=4*(l/3)+(l%3)+!!(l%3);
 char *encoded=(char*)malloc(4*(le/4)+4*!!(le%4)+1);
 int i,j,k;

 for(i=j=0;i<(l+3);i+=3,j+=4)
   {
    unsigned long s=0;

    for(k=0;k<3;k++)
       if((i+k)<l)
          s|=((unsigned long)str[i+k]&0xff)<<(16-8*k);

    for(k=0;k<4;k++)
       if((j+k)<le)
          encoded[j+k]=(char)((s>>(18-6*k))&0x3f);
   }

 for(j=0;j<le;j++)
    encoded[j]=base64[(int)encoded[j]];
 for(;j%4;j++)
    encoded[j]='=';
 encoded[j]=0;

 return(encoded);
}


/*++++++++++++++++++++++++++++++++++++++
  Make the input string safe to output as HTML ( not < > & " ).

  char* HTMLString Returns a safe HTML string.

  const char* c A non-safe HTML string.

  int nbsp Use a non-breaking space in place of normal ones.
  ++++++++++++++++++++++++++++++++++++++*/

char* HTMLString(const char* c,int nbsp)
{
 int i=0,j=0,len=256-5;              /* 5 is the longest possible inserted amount */
 char* ret=(char*)malloc(257);

 do
   {
    for(;j<len && c[i];i++)
       switch(c[i])
         {
         case '<':
          ret[j++]='&';
          ret[j++]='l';
          ret[j++]='t';
          ret[j++]=';';
          break;
         case '>':
          ret[j++]='&';
          ret[j++]='g';
          ret[j++]='t';
          ret[j++]=';';
          break;
         case '"':
          ret[j++]='&';
          ret[j++]='q';
          ret[j++]='u';
          ret[j++]='o';
          ret[j++]='t';
          ret[j++]=';';
          break;
         case '&':
          ret[j++]='&';
          ret[j++]='a';
          ret[j++]='m';
          ret[j++]='p';
          ret[j++]=';';
          break;
         case ' ':
          if(nbsp)
            {
             ret[j++]='&';
             ret[j++]='n';
             ret[j++]='b';
             ret[j++]='s';
             ret[j++]='p';
             ret[j++]=';';
             break;
            }
         /* else fallthrough */
         default:
          ret[j++]=c[i];
         }

    if(c[i])                 /* Not finished */
      {
       ret=(char*)realloc((void*)ret,len+256+5);
       len+=256;
      }
   }
 while(c[i]);

 ret[j]=0;

 return(ret);
}
